<?php
require_once __DIR__ . '/class.php';
require_once __DIR__ . '/vendor/autoload.php';
require_once __DIR__ . '/phpseclib-master/vendor/autoload.php';


use Aws\ElasticLoadBalancing\ElasticLoadBalancingClient;
use Aws\AutoScaling\AutoScalingClient;
use Aws\CloudWatch\CloudWatchClient;
use Aws\Ec2\Ec2Client;
use Aws\Rds\RdsClient;
use Aws\Sns\SnsClient;

use phpseclib\Crypt\RSA;
use phpseclib\Net\SSH2;
use phpseclib\Net\SCP;

class DemoApi
{

    private $ec2Client, $asClient, $cwClient, $elbClient, $rdsClient, $snsClient,
            $internetGatewayID, $vpcID, $subnetAID, $subnetBID, $securityGroupID,
            $ec2InstanceID, $ec2InstanceIP, $ec2ImageID, $ssh,
            $haproxyInstanceID, $haproxyInstanceIP, $dbInstanceIP,
            $scaleInPolicyARN, $scaleOutPolicyARN, $subscriptionARN, $topicARN;

    public function __construct()
    {
        $params = array('region'  => AWS_REGION,
                        'version' => AWS_VERSION,
                        'credentials' => [
                            'key'    => 'J0HNdOEJOHND0EJOHNDO',
                            'secret' => 'J0hndOEjOHND0eJoHNDOj0HNd0EJOHND0EJOHNDo',
                        ]);

        $this->ec2Client = new Ec2Client($params);
        $this->cwClient  = new CloudWatchClient($params);
        $this->asClient  = new AutoScalingClient($params);
        $this->elbClient = new ElasticLoadBalancingClient($params);
        $this->rdsClient = new RdsClient($params);
        $this->snsClient = new SnsClient($params);
    }

    public function createInternetGateway()
    {
        try {
            $result = $this->ec2Client->createInternetGateway([]);
            $this->internetGatewayID = $result->get('InternetGateway')['InternetGatewayId'];
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createVPC()
    {
        try {
            $result = $this->ec2Client->createVpc(['CidrBlock' => VPC_CIDR]);
            $this->vpcID = $result->get('Vpc')['VpcId'];
            $this->ec2Client->attachInternetGateway([
                'InternetGatewayId' => $this->internetGatewayID,
                'VpcId' => $this->vpcID
            ]);
            $this->ec2Client->modifyVpcAttribute([
                'EnableDnsHostnames' => ['Value' => true],
                'VpcId' => $this->vpcID
            ]);
            $this->ec2Client->modifyVpcAttribute([
                'EnableDnsSupport' => ['Value' => true],
                'VpcId' => $this->vpcID
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createSubnets()
    {
        try{
            $result = $this->ec2Client->createSubnet([
                'AvailabilityZone'  => AWS_REGION . 'a',
                'CidrBlock'         => SUBNET_A_CIDR,
                'VpcId'             => $this->vpcID
            ]);
            $this->subnetAID = $result->get('Subnet')['SubnetId'];
            $this->ec2Client->modifySubnetAttribute([
                'MapPublicIpOnLaunch' => ['Value' => true],
                'SubnetId'            => $this->subnetAID
            ]);

            $result = $this->ec2Client->createSubnet([
                'AvailabilityZone'  => AWS_REGION . 'b',
                'CidrBlock'         => SUBNET_B_CIDR,
                'VpcId'             => $this->vpcID
            ]);
            $this->subnetBID = $result->get('Subnet')['SubnetId'];
            $this->ec2Client->modifySubnetAttribute([
                'MapPublicIpOnLaunch' => ['Value' => true],
                'SubnetId' => $this->subnetBID
            ]);

            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function prepareRouteTable()
    {
        try {
            $result = $this->ec2Client->describeRouteTables([
                'Filters' => [
                    [
                        'Name'   => 'vpc-id',
                        'Values' => [$this->vpcID]
                    ]
                ]
            ]);
            $routeTableID = $result->get('RouteTables')[0]['RouteTableId'];

            $this->ec2Client->createRoute([
                'DestinationCidrBlock'  => '0.0.0.0/0',
                'GatewayId'             => $this->internetGatewayID,
                'RouteTableId'          => $routeTableID
            ]);

            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createSecurityGroup()
    {
        try {
            $result = $this->ec2Client->createSecurityGroup([
                'Description'   => 'API Security Group for SSH, HTTP and MySQL connections.',
                'GroupName'     => NAME_SECURITY_GROUP,
                'VpcId'         => $this->vpcID
            ]);
            $this->securityGroupID = $result->get('GroupId');

            $this->ec2Client->authorizeSecurityGroupIngress([
                'GroupId'       => $this->securityGroupID,
                'IpProtocol'    => 'TCP',
                'ToPort'        => 22,
                'FromPort'      => 22,
                'CidrIp'        => '0.0.0.0/0'
            ]);
            $this->ec2Client->authorizeSecurityGroupIngress([
                'GroupId'       => $this->securityGroupID,
                'IpProtocol'    => 'TCP',
                'ToPort'        => 80,
                'FromPort'      => 80,
                'CidrIp'        => '0.0.0.0/0'
            ]);
            $this->ec2Client->authorizeSecurityGroupIngress([
                'GroupId'       => $this->securityGroupID,
                'IpProtocol'    => 'TCP',
                'ToPort'        => 3306,
                'FromPort'      => 3306,
                'CidrIp'        => '0.0.0.0/0'
            ]);
            return $this;
        }  catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createKeyPair()
    {
        try {
            $result = $this->ec2Client->createKeyPair([
                'KeyName' => NAME_KEY_PAIR,
            ]);
            $key = $result->get('KeyMaterial');
            $filename = __DIR__ . '/' . NAME_KEY_PAIR . '.pem';
            file_put_contents($filename, $key);
            chmod($filename, 0400);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createDBSubnetGroup()
    {
        try {
            $this->rdsClient->createDBSubnetGroup([
                'DBSubnetGroupDescription'  => 'API security group for RDS connections.',
                'DBSubnetGroupName'         => NAME_DB_SUBNET_GROUP,
                'SubnetIds'                 => [$this->subnetAID, $this->subnetBID],
                'Tags'                      => [
                    [
                        'Key'   => 'Name',
                        'Value' => NAME_DB_SUBNET_GROUP,
                    ]
                ]
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createDBInstance()
    {
        try {
            $result = $this->rdsClient->createDBInstance([
                'AllocatedStorage'          => DB_INSTANCE_STORAGE,
                'AutoMinorVersionUpgrade'   => true,
                'AvailabilityZone'          => AWS_REGION . 'a',
                'BackupRetentionPeriod'     => 35,
                'DBInstanceClass'           => DB_INSTANCE_TYPE,
                'DBInstanceIdentifier'      => DB_INSTANCE_NAME,
                'DBName'                    => DB_NAME,
                'DBSubnetGroupName'         => NAME_DB_SUBNET_GROUP,
                'Engine'                    => DB_INSTANCE_ENGINE,
                'EngineVersion'             => DB_INSTANCE_ENGINE_VERSION,
                'LicenseModel'              => 'general-public-license',
                'MasterUserPassword'        => DB_PASSWORD,
                'MasterUsername'            => DB_USERNAME,
                'MultiAZ'                   => false,
                'Port'                      => 3306,
                'PubliclyAccessible'        => true,
                'StorageType'               => 'gp2',
                'VpcSecurityGroupIds'       => [$this->securityGroupID]
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createEC2Instance()
    {
        try {
            $result = $this->ec2Client->runInstances([
                'BlockDeviceMappings'               => [
                    [
                        'DeviceName'    => '/dev/xvda',
                        'Ebs'           => [
                            'DeleteOnTermination'   => true,
                            'SnapshotId'            => 'snap-2974047b',
                            'VolumeSize'            => 8,
                            'VolumeType'            => 'gp2',
                        ]
                    ]
                ],
                'ImageId'                           => EC2_AMI_ID,
                'InstanceInitiatedShutdownBehavior' => 'stop',
                'InstanceType'                      => EC2_INSTANCE_TYPE,
                'KeyName'                           => NAME_KEY_PAIR,
                'MaxCount'                          => 1,
                'MinCount'                          => 1,
                'Monitoring'                        => ['Enabled' => false],
                'SecurityGroupIds'                  => [$this->securityGroupID],
                'SubnetId'                          => $this->subnetAID
            ]);
            $this->ec2InstanceID = $result->get('Instances')[0]['InstanceId'];

            $this->ec2Client->createTags([
                'Resources' => [$this->ec2InstanceID],
                'Tags'      => [
                    [
                        'Key'   => 'Name',
                        'Value' => NAME_EC2_INSTANCE
                    ]
                ]
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        } 
    }

    public function createHAProxyInstance()
    {
        try {
            $result = $this->ec2Client->runInstances([
                'BlockDeviceMappings'               => [
                    [
                        'DeviceName'    => '/dev/xvda',
                        'Ebs'           => [
                            'DeleteOnTermination'   => true,
                            'SnapshotId'            => 'snap-2974047b',
                            'VolumeSize'            => 8,
                            'VolumeType'            => 'gp2',
                        ]
                    ]
                ],
                'ImageId'                           => EC2_AMI_ID,
                'InstanceInitiatedShutdownBehavior' => 'stop',
                'InstanceType'                      => EC2_INSTANCE_TYPE,
                'KeyName'                           => NAME_KEY_PAIR,
                'MaxCount'                          => 1,
                'MinCount'                          => 1,
                'Monitoring'                        => ['Enabled' => false],
                'SecurityGroupIds'                  => [$this->securityGroupID],
                'SubnetId'                          => $this->subnetAID
            ]);
            $this->haproxyInstanceID = $result->get('Instances')[0]['InstanceId'];

            $this->ec2Client->createTags([
                'Resources' => [$this->haproxyInstanceID],
                'Tags' => [
                    [
                        'Key' => 'Name',
                        'Value' => NAME_HAPROXY_INSTANCE
                    ]
                ]
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function getDBEndpoint()
    {
        $count = 0;
        do {
            if ($count++) sleep(INSTANCE_CHECK_TIMEOUT);
            $result = $this->rdsClient->describeDBInstances([
                'DBInstanceIdentifier' => DB_INSTANCE_NAME
            ]);
            $this->dbInstanceIP = $result->get('DBInstances')[0]['Endpoint']['Address'];
        } while (!$this->dbInstanceIP && $count < INSTANCE_CHECK_COUNT);

        return $this;
    }

    public function prepareDBInstance()
    {
        try {
            $this->editFile('/files/users.sql',
                ['{HAPROXY_INSTANCE_IP}', '{DB_WP_USERNAME}',
                 '{DB_WP_PASSWORD}', '{DB_NAME}'],
                [$this->haproxyInstanceIP, DB_WP_USERNAME,
                 DB_WP_PASSWORD, DB_NAME]);

            $command = 'mysql -h ' . $this->dbInstanceIP . ' -P 3306 '
                     . '--user=' . DB_USERNAME . ' --password=' . DB_PASSWORD
                     . ' ' . DB_NAME . ' < ' . __DIR__ . '/files/users.sql';
            shell_exec($command);
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function getInstanceIP($id)
    {
        $count = 0;
        do {
            if ($count++) sleep(INSTANCE_CHECK_TIMEOUT);
            $result = $this->ec2Client->describeInstances([
                'InstanceIds' => [$id]
            ]);
            $ip = $result->get('Reservations')[0]['Instances'][0]['PublicIpAddress'];
        } while (!$ip && $count < INSTANCE_CHECK_COUNT);
        return $ip;
    }

    private function getInstanceStatus($id)
    {
        $count = 0;
        do {
            if ($count++) sleep(INSTANCE_CHECK_TIMEOUT);
            $result = $this->ec2Client->describeInstanceStatus([
                'InstanceIds' => [$id]
            ]);
            $status = $result->get('InstanceStatuses')[0]['InstanceStatus']['Status'];
        } while ($status != 'ok' && $count < INSTANCE_CHECK_COUNT);

        return $status;
    }

    public function prepareHAProxyInstance()
    {
        try {
            $this->haproxyInstanceIP = $this->getInstanceIP($this->haproxyInstanceID);
            $this->getInstanceStatus($this->haproxyInstanceID);
            $this->connectToInstance($this->haproxyInstanceIP);
            $this->ssh->exec('mkdir files');
            $this->executeCommands('/files/haproxy_install.sh');
            $this->waitExecution();

            $this->connectToInstance($this->haproxyInstanceIP);
            $this->editHAProxyConfig();
            $files = array(
                __DIR__ . '/files/haproxy.conf' => '/etc/rsyslog.d/haproxy.conf',
                __DIR__ . '/files/haproxy.cfg'  => '/etc/haproxy/haproxy.cfg',
                __DIR__ . '/files/rsyslog.conf' => '/home/ec2-user/rsyslog.conf'
            );

            $this->sendFiles($files);
            $this->executeCommands('/files/haproxy_services_start.sh');
            $this->waitExecution();
            $this->disconnectFromInstance();

            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function editHAProxyConfig()
    {
        (new SCP($this->ssh))->get('/etc/rsyslog.conf', __DIR__ . '/files/rsyslog.conf');
        $this->editFile('/files/rsyslog.conf',["# Provides UDP syslog reception"],
                        ["# Provides UDP syslog reception\r\n\$ModLoad imudp\r\n"
                        . "\$UDPServerAddress 127.0.0.1\r\n\$UDPServerRun 514"])
             ->editFile('/files/haproxy.cfg', ["server api-db-instance"],
                        ["server api-db-instance " . $this->dbInstanceIP . ":3306 check"]);
        return $this;
    }

    public function prepareEC2Instance()
    {
        try {
            $this->editWordpressConfig();
            $this->ec2InstanceIP = $this->getInstanceIP($this->ec2InstanceID);
            $this->getInstanceStatus($this->ec2InstanceID);
            $this->connectToInstance($this->ec2InstanceIP);
            $this->ssh->exec('mkdir files');
            $this->executeCommands('/files/lemp_install.sh');
            $this->waitExecution();

            $this->connectToInstance($this->ec2InstanceIP);
            $files = array(
                __DIR__ . '/files/nginx.conf'    => '/etc/nginx/nginx.conf',
                __DIR__ . '/files/www.conf'      => '/etc/php-fpm.d/www.conf',
                __DIR__ . '/files/wp-config.php' => '/var/www/wp-config.php'
            );
            $this->sendFiles($files);
            $this->executeCommands('/files/lemp_services_start.sh');
            $this->waitExecution();
            $this->disconnectFromInstance();

            return $this;
        }  catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function editWordpressConfig()
    {
        $this->editFile('/files/wp-config.php',
            ["define('DB_HOST', 'localhost');",
             "define('FTP_HOST', '');"],
            ["define('DB_HOST', '" . $this->haproxyInstanceIP . "');",
             "define('FTP_HOST', '" . SCRIPT_HOST . "');"]
        );
        return $this;
    }

    public function createEC2Image()
    {
        try {
            $result = $this->ec2Client->createImage([
                'Description'   => 'API Image: LEMP + WP',
                'InstanceId'    => $this->ec2InstanceID,
                'Name'          => 'api-ec2-image'
            ]);
            $this->ec2ImageID = $result->get('ImageId');
            return $this;
        }  catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createLoadBalancer()
    {
        try {
            $this->elbClient->createLoadBalancer([
                'Listeners' => [
                    [
                        'InstancePort'      => 80,
                        'InstanceProtocol'  => 'HTTP',
                        'LoadBalancerPort'  => 80,
                        'Protocol'          => 'HTTP',
                    ]
                ],
                'LoadBalancerName'  => NAME_LOAD_BALANCER,
                'Scheme'            => 'internet-facing',
                'SecurityGroups'    => [$this->securityGroupID],
                'Subnets'           => [$this->subnetAID],
                'Tags'              => [
                    [
                        'Key'   => 'Name',
                        'Value' => NAME_LOAD_BALANCER
                    ]
                ]
            ]);
            $this->elbClient->configureHealthCheck([
                'HealthCheck'       => [
                    'HealthyThreshold'      => 10,
                    'Interval'              => 30,
                    'Target'                => 'HTTP:80/',
                    'Timeout'               => 5,
                    'UnhealthyThreshold'    => 2,
                ],
                'LoadBalancerName'  => NAME_LOAD_BALANCER,
            ]);
            $this->elbClient->registerInstancesWithLoadBalancer([
                'Instances'         => [
                    ['InstanceId'   => $this->ec2InstanceID]
                ],
                'LoadBalancerName'  => NAME_LOAD_BALANCER
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createLaunchConfiguration()
    {
        try {
            $this->asClient->createLaunchConfiguration([
                'AssociatePublicIpAddress'  => true,
                'EbsOptimized'              => false,
                'ImageId'                   => $this->ec2ImageID,
                'InstanceMonitoring'        => ['Enabled' => false],
                'InstanceType'              => EC2_INSTANCE_TYPE,
                'KeyName'                   => NAME_KEY_PAIR,
                'LaunchConfigurationName'   => NAME_LAUNCH_CONFIGURATION,
                'SecurityGroups'            => [$this->securityGroupID],
                'SpotPrice'                 => SPOT_INSTANCE_PRICE,
                'Tags'                      => [
                    [
                        'Key'   => 'Name',
                        'Value' => NAME_LAUNCH_CONFIGURATION
                    ]
                ]
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createAutoScalingGroup()
    {
        try {
            $this->asClient->createAutoScalingGroup([
                'AvailabilityZones'         => [AWS_REGION . 'a'],
                'AutoScalingGroupName'      => NAME_AUTO_SCALING_GROUP,
                'DesiredCapacity'           => 0,
                'LaunchConfigurationName'   => NAME_LAUNCH_CONFIGURATION,
                'LoadBalancerNames'         => [NAME_LOAD_BALANCER],
                'MaxSize'                   => 3,
                'MinSize'                   => 0,
                'Tags'                      => [
                    [
                        'Key'   => 'Name',
                        'Value' => NAME_AUTO_SCALING_GROUP
                    ]
                ],
                'VPCZoneIdentifier'         => $this->subnetAID
            ]);
            return $this;
        }  catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createScaleInPolicy()
    {
        try {
            $result = $this->asClient->putScalingPolicy([
                'AdjustmentType'        => 'ChangeInCapacity',
                'AutoScalingGroupName'  => NAME_AUTO_SCALING_GROUP,
                'PolicyName'            => 'api-scalein-policy',
                'PolicyType'            => 'SimpleScaling',
                'ScalingAdjustment'     => 1
            ]);
            $this->scaleInPolicyARN = $result->get('PolicyARN');
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createScaleOutPolicy()
    {
        try {
            $result = $this->asClient->putScalingPolicy([
                'AdjustmentType'        => 'ChangeInCapacity',
                'AutoScalingGroupName'  => NAME_AUTO_SCALING_GROUP,
                'PolicyName'            => 'api-scaleout-policy',
                'PolicyType'            => 'SimpleScaling',
                'ScalingAdjustment'     => -1
            ]);
            $this->scaleOutPolicyARN = $result->get('PolicyARN');
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createSNSNotification()
    {
        try {
            $result = $this->snsClient->createTopic([
                'Name' => 'api-sns-topic'
            ]);
            $this->topicARN = $result->get('TopicArn');
            $this->snsClient->subscribe(array(
                'TopicArn' => $this->topicARN,
                'Protocol' => 'http',
                'Endpoint' => 'http://' . SCRIPT_HOST . '/action.php',
            ));
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function getSNSSubscription($message)
    {
        try {
            $result = $this->snsClient->confirmSubscription(array(
                    'Token'    => $message['Token'],
                    'TopicArn' => $message['TopicArn']
                ));
            $this->subscriptionARN = $result->get('SubscriptionArn');
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createAddCapacityAlarm()
    {
        try {
            $this->cwClient->putMetricAlarm([
                'AlarmName'          => ALARM_ADD_NAME,
                'AlarmDescription'   => 'Add instance when CPU Utilization is '
                                        . 'grater or equal than threshold.',
                'AlarmActions'       => [
                    $this->scaleInPolicyARN,
                    $this->topicARN
                ],
                'ComparisonOperator' => ALARM_ADD_COMPARISON_OPERATOR,
                'Dimensions'         => [
                    [
                        'Name'  => 'InstanceId',
                        'Value' => $this->ec2InstanceID
                    ]
                ],
                'MetricName'        => ALARM_ADD_METRIC_NAME,
                'Namespace'         => ALARM_ADD_NAMESPACE,
                'EvaluationPeriods' => ALARM_ADD_EVALUATION_PERIODS,
                'Period'            => ALARM_ADD_PERIOD,
                'Statistic'         => ALARM_ADD_STATISTIC,
                'Threshold'         => ALARM_ADD_THRESHOLD
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function createRemoveCapacityAlarm()
    {
        try {
            $result = $this->cwClient->putMetricAlarm([
                'AlarmName'          => ALARM_REMOVE_NAME,
                'AlarmDescription'   => 'Remove instance when CPU Utilization is '
                                        . 'lower than threshold.',
                'AlarmActions'       => [
                    $this->scaleOutPolicyARN,
                    $this->topicARN
                ],
                'ComparisonOperator' => ALARM_REMOVE_COMPARISON_OPERATOR,
                'Dimensions'         => [
                    [
                        'Name'  => 'InstanceId',
                        'Value' => $this->ec2InstanceID
                    ]
                ],
                'MetricName'        => ALARM_REMOVE_METRIC_NAME,
                'Namespace'         => ALARM_REMOVE_NAMESPACE,
                'EvaluationPeriods' => ALARM_REMOVE_EVALUATION_PERIODS,
                'Period'            => ALARM_REMOVE_PERIOD,
                'Statistic'         => ALARM_REMOVE_STATISTIC,
                'Threshold'         => ALARM_REMOVE_THRESHOLD
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function runCronJob($jobName)
    {
        $minute = intval(date('i')) + 1;
        $job = "$minute * * * * /usr/bin/php /var/www/$jobName.php "
             . "> /var/www/$jobName.log" . PHP_EOL;
        $filename = __DIR__ . '/crontab.txt';
        file_put_contents($filename, $job);
        shell_exec("sudo crontab -u ec2-user $filename 2>&1");
        unlink($filename);
        $result = shell_exec('sudo crontab -u ec2-user -l 2>&1');
        $this->logAction(__FUNCTION__, "Running cron job", $result);
    }

    public function addDBReplica()
    {
        try {
            $id = DB_INSTANCE_REPLICA_NAME . '-' . time();
            $prefix = $this->getDBPrefix();
            $this->createDBReplica($id);

            $count = 0;
            do {
                if ($count++) sleep(INSTANCE_CHECK_TIMEOUT);
                $result = $this->rdsClient->describeDBInstances([
                    'DBInstanceIdentifier' => $id
                ]);
                $status = $result->get('DBInstances')[0]['DBInstanceStatus'];
            } while ($status != 'available' && $count < INSTANCE_CHECK_COUNT);

            $this->addRecordToHAProxy($id, $prefix);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    public function removeDBReplica()
    {
        try {
            $id = $this->getLastReplica();
            $prefix = $this->getDBPrefix();
            $this->deleteDBReplica($id);
            $this->removeRecordFromHAProxy($id, $prefix);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function createDBReplica($id)
    {
        try {
            $this->rdsClient->createDBInstanceReadReplica([
                'AutoMinorVersionUpgrade'       => true,
                'AvailabilityZone'              => AWS_REGION . 'a',
                'CopyTagsToSnapshot'            => true,
                'DBInstanceClass'               => DB_INSTANCE_TYPE,
                'DBInstanceIdentifier'          => $id,
                'Port'                          => 3306,
                'PubliclyAccessible'            => true,
                'SourceDBInstanceIdentifier'    => DB_INSTANCE_NAME,
                'StorageType'                   => 'gp2'
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function deleteDBReplica($id)
    {
        try {
            $this->rdsClient->deleteDBInstance([
                'DBInstanceIdentifier'  => $id,
                'SkipFinalSnapshot'     => true
            ]);
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function getLastReplica()
    {
        try {
            $result = $this->rdsClient->describeDBInstances([
                'DBInstanceIdentifier' => DB_INSTANCE_NAME
            ]);
            $replicas = $result->get('DBInstances')[0]['ReadReplicaDBInstanceIdentifiers'];
            
            foreach ($replicas as $replica) {
                $result = $this->rdsClient->describeDBInstances([
                    'DBInstanceIdentifier' => $replica
                ]);
                $status = $result->get('DBInstances')[0]['DBInstanceStatus'];
                if ($status == 'available') {
                    $lastReplica = $replica;
                    break;
                }
            }
            return $lastReplica;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function getDBPrefix()
    {
        $this->getDBEndpoint();
        $start = strpos($this->dbInstanceIP, '.');
        return substr($this->dbInstanceIP, $start);
    }

    private function getInstanceID($name)
    {
        try {
            $result = $this->ec2Client->describeInstances([
                'Filters' => [
                    [
                        'Name' => 'tag-key',
                        'Values' => ['Name']
                    ],
                    [
                        'Name' => 'tag-value',
                        'Values' => [$name]
                    ]
                ]
            ]);
            $id = $result->get('Reservations')[0]['Instances'][0]['InstanceId'];
            return $id;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function addRecordToHAProxy($id, $prefix)
    {
        try {   
            $record  = "\nserver " . $id . " " . $id . $prefix . ":3306 check";        
            $content = file_get_contents(__DIR__ . '/files/haproxy.cfg');
            $start   = strrpos($content, ":3306 check") + strlen(":3306 check");
            $content = substr_replace($content, $record, $start, 0);
            file_put_contents(__DIR__ . '/files/haproxy.cfg', $content);
            $this->changeHAProxyConfig();
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function removeRecordFromHAProxy($id, $prefix)
    {
        try {
            $record  = "\nserver " . $id . " " . $id . $prefix . ":3306 check";
            $this->editFile('/files/haproxy.cfg', [$record], ['']);
            $this->changeHAProxyConfig();
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function changeHAProxyConfig()
    {
        try {
            $haproxyID = $this->getInstanceID(NAME_HAPROXY_INSTANCE);
            $haproxyIP = $this->getInstanceIP($haproxyID);
            $this->connectToInstance($haproxyIP);
            $files = array(
                __DIR__ . '/files/haproxy.cfg' => '/etc/haproxy/haproxy.cfg'
            );
            $this->sendFiles($files);
            $result = $this->ssh->exec('sudo service haproxy restart');
            $this->logAction(__FUNCTION__, "Restarting HAProxy", $result);
            $this->disconnectFromInstance();
            return $this;
        } catch (Exception $e) {
            $this->logException(__FUNCTION__, $e->getMessage());
            return false;
        }
    }

    private function connectToInstance($ip)
    {
        $key = new RSA();
        $key->load(file_get_contents(__DIR__ . '/' . NAME_KEY_PAIR . '.pem'));
        $this->ssh = new SSH2($ip);
        if (!$this->ssh->login(SSH_USERNAME, $key)) {
            throw new Exception('SSH login has been failed.');
        }
        return $this;
    }

    private function disconnectFromInstance()
    {
        $this->ssh->exec('exit');
        unset($this->ssh);
        return $this;
    }

    private function sendFiles($files)
    {
        $scp = new SCP($this->ssh);
        foreach ($files as $local => $remote) {
            $result = $scp->put($remote, $local, SCP::SOURCE_LOCAL_FILE);
            $this->logAction(__FUNCTION__, "Sending file $remote", $result);
        }
        return $this;
    }

    private function executeCommands($script)
    {
        $remoteScript = INSTANCE_HOME_DIR . $script;
        $this->sendFiles([__DIR__ . $script => $remoteScript]);
        $result = $this->ssh->exec('sh ' . $remoteScript);
        $this->logAction(__FUNCTION__, "Execution of $remoteScript", $result);
        return $this;
    }

    private function editFile($filename, $old, $new)
    {
        $content = file_get_contents(__DIR__ . $filename);
        foreach ($old as $no => $line) {
            $content = str_replace($line, $new[$no], $content);
        }
        file_put_contents(__DIR__ . $filename, $content);
        return $this;
    }

    private function logException($function, $exception)
    {
        $log = "$function : $exception\r\n\r\n";
        file_put_contents(__DIR__ . '/error.log', $log, FILE_APPEND);
        return $this;
    }

    private function logAction($function, $action, $message)
    {
        $log = "$function : $action\r\n$message\r\n\r\n";
        file_put_contents(__DIR__ . '/script.log', $log, FILE_APPEND);
    }

    private function logMessage($message)
    {
        file_put_contents(__DIR__ . '/message.log', $message . "\r\n\r\n", FILE_APPEND);
    }

    public function removeCronJob()
    {
        shell_exec('sudo crontab -u ec2-user -r 2>&1');
    }

    private function waitExecution()
    {
        sleep(120);
    }
}
?>